# 3rd-party
from lxml import etree
from lxml.builder import E

# package modules
from jnpr.junos import jxml as JXML
from jnpr.junos.cfg.resource import Resource, P_JUNOS_EXISTS, P_JUNOS_ACTIVE


class TemplateResource(Resource):

    # -----------------------------------------------------------------------
    # PROPERTIES
    # -----------------------------------------------------------------------

    @property
    def active(self):
        """
          returns :True: if every named element in the XML is active
          returns :False: otherwise
        """
        _active = True

        for name in self._name:
            if not self.has[P_JUNOS_ACTIVE + '_' + name]:
                _active = False
                break

        return _active

    @property
    def exists(self):
        """
          returns :True: if every named element in the XML exists
          returns :False: otherwise
        """
        _exists = True

        for name in self._name:
            if not self.has[P_JUNOS_EXISTS + '_' + name]:
                _exists = False
                break

        return _exists

    # -------------------------------------------------------------------------
    # activate
    # -------------------------------------------------------------------------

    def activate(self, **kvargs):
        """
          write config to activate resource; i.e. "activate ..."
        """
        #! removed check on active due to multi-key template
        #@@@ revisit

        names = kvargs if len(kvargs) else self._name

        xml = self._xml_template_active(JXML.ACTIVATE, names)
        rsp = self._r_config_write_xml(xml)

        for name in names:
            self.has[P_JUNOS_ACTIVE + '_' + name] = True

        return True

    # -------------------------------------------------------------------------
    # deactivate
    # -------------------------------------------------------------------------

    def deactivate(self, **kvargs):
        """
          write config to deactivate resource, i.e. "deactivate ..."
        """
        #! removed check on active due to multi-key template
        #@@@ revisit

        names = kvargs if len(kvargs) else self._name

        xml = self._xml_template_active(JXML.DEACTIVATE, names)
        rsp = self._r_config_write_xml(xml)

        for name in names:
            self.has[P_JUNOS_ACTIVE + '_' + name] = False

        return True

    # -------------------------------------------------------------------------
    # delete
    # -------------------------------------------------------------------------

    def delete(self, **kvargs):
        """
          remove configuration from Junos device
        """
        # ! removed check on exists due to multi-key template
        # ! @@ revisit

        names = kvargs if len(kvargs) else self._name

        xml = self._xml_template_delete(names)
        rsp = self._r_config_write_xml(xml)

        for name in names:
            self.has[P_JUNOS_EXISTS + '_' + name] = False

        return True

    # -------------------------------------------------------------------------
    # rename
    # -------------------------------------------------------------------------

    def rename(self, new_name=None, **kvargs):
        """
          Not supported (for now) - raise RuntimeError
        """
        raise RuntimeError(
            "rename not supported on %s" %
            self.__class__.__name__)

    # -------------------------------------------------------------------------
    # reorder
    # -------------------------------------------------------------------------

    def reorder(self, **kvargs):
        """
          Not supported - raises RuntimeError
        """
        raise RuntimeError(
            "reorder not supported on %s" %
            self.__class__.__name__)

    # -----------------------------------------------------------------------
    # ~private~ OVERLOADING
    # -----------------------------------------------------------------------

    def _r_config_read_xml(self):
        """
        ~! OVERLOADS :Resource: !~
          read the resource config from the Junos device
        """
        return self._junos.rpc.get_config(self._xml_template_read())

    def _xml_build_change(self):
        """
        ~! OVERLOADS :Resource !~
          run the templater to produce the XML for change.  this
          method is invoked by the parent :Resource: during the
          invocation of the :write():
        """
        return self._xml_template_write()

    # -----------------------------------------------------------------------
    # ~private~ resource helper methods
    # -----------------------------------------------------------------------

    def _r_has_init(self):
        """
          initializes the :has: attribute
        """
        self.has.clear()
        for name in self._name:
            self.has[P_JUNOS_EXISTS + '_' + name] = False
            self.has[P_JUNOS_ACTIVE + '_' + name] = False

    def _r_set_active(self, my_props, value):
        """
          sets each of the :has: name P_JUNOS_ACTIVE
          properties to :value:
        """
        for name in self._name:
            my_props[P_JUNOS_ACTIVE + '_' + name] = value

    def _r_set_exists(self, my_props, value):
        """
          sets each of the name P_JUNOS_ACTIVE
          properties to in the :my_props: to :value:
        """
        for name in self._name:
            my_props[P_JUNOS_EXISTS + '_' + name] = value

    def _r_has_xml_status(self, xml_ele, to_py):
        """
          set the P_JUNOS_EXISTS and P_JUNOS_ACTIVE property
          states for each item in the resource name as extracted
          from the XML configuration retrieved from the Junos device
        """
        for name in self._xpath_names:
            if xml_ele[name] is not None:
                to_py[P_JUNOS_EXISTS + '_' + name] = True
                active = False if 'inactive' in xml_ele[name].attrib else True
                to_py[P_JUNOS_ACTIVE + '_' + name] = active
            else:
                to_py[P_JUNOS_EXISTS + '_' + name] = False
                to_py[P_JUNOS_ACTIVE + '_' + name] = False

    def _r_template_write_vars(self):
        """
          This method is called when the XML config-write template
          is being formed.  by default the simplest case is to combine
          all of the variables together.  more complicated template
          resources can overload this method to do what they want.
        """
        t_vars = dict(self._name)
        t_vars.update(self.has)
        t_vars.update(self.should)
        return t_vars

    # -----------------------------------------------------------------------
    # ~private~ methods used to create XML structures from templates
    # -----------------------------------------------------------------------

    def _xml_template_active(self, cmd, names):
        """
          create an XML structure that will activate/deactivate
          the named elements in the resource configuration
        """

        xml = self._xml_template_names_only()
        for name, xpath in self._xpath_names.items():
            if name not in names:
                continue
            ele = xml.find(xpath)
            ele.attrib.update(cmd)
        return xml

    def _xml_template_delete(self, names):
        """
          creates an XML structure that will delete
          the named elements in the resource configuration
        """

        xml = self._xml_template_names_only()
        for name, xpath in self._xpath_names.items():
            if name not in names:
                continue
            ele = xml.find(xpath)
            ele.attrib.update(JXML.DEL)
        return xml

    def _xml_template_read(self):
        """
          create an XML structure from template that
          will be used to read the resource configuration
          from the Junos device
        """
        t = self._j2_ldr.get_template(self._j2_rd)
        return etree.XML(t.render(self._name))

    def _xml_template_names_only(self):
        """
          create an XML structure from template that will
          be used to read only the resource name values
          from the Junos device
        """
        t = self._j2_ldr.get_template(self._j2_rd)
        return etree.XML(t.render(self._name, NAMES_ONLY=True))

    def _xml_template_write(self):
        t_vars = self._r_template_write_vars()
        t = self._j2_ldr.get_template(self._j2_wr)
        return etree.XML(t.render(t_vars))



# 3rd-party modules
from lxml.builder import E

# module packages
from jnpr.junos.cfg import Resource
from jnpr.junos import jxml as JXML


class NatProxyArp(Resource):

    """
    [edit security nat proxy-arp interface <if_name> address <ip_prefix>]

    Resource namevar:
      tuple(if_name, ip_prefix)

    Description:
      This resource allows you to add/remove proxy-arp entries for NAT.  At
      this time, there are no managed properties, so you can simply add or
      remove entries by the name tuple(if_name, ip_prefix)

      For example, to select an entry directly:

        entry = NatProxyArp(jdev, ('reth0.213','198.18.11.5'))

      Or using the bind mechanism:

        jdev.bind(parp=NatProxyArp)
        entry = jdev.parp[('reth0.213', '198.18.11.5')]

      To create it, you need to use the 'touch' option when invoking
      write() since there are no properites for proxy-arp entries

        if not entry.exists:
          entry.write(touch=True)

      And to remove the same entry:

        entry.delete()
    """

    def _xml_at_top(self):
        return E.security(E.nat(
            E('proxy-arp',
              E.interface(E.name(self._name[0]),
                          E.address(E.name(self._name[1]))
                          )
              )
        ))

    # -----------------------------------------------------------------------
    # OVERLOADS
    # -----------------------------------------------------------------------

    def rename(self, name):
        """ UNSUPPORTED """
        raise RuntimeError(
            "Unsupported for Resource: %s" %
            self.__class__.__name__)

    # -----------------------------------------------------------------------
    # XML read
    # -----------------------------------------------------------------------

    def _xml_at_res(self, xml):
        return xml.find('.//proxy-arp/interface')

    def _xml_to_py(self, as_xml, to_py):
        Resource._r_has_xml_status(as_xml, to_py)

    # -----------------------------------------------------------------------
    # Resource List, Catalog
    # -- only executed by 'manager' resources
    # -----------------------------------------------------------------------

    def _r_list(self):
        raise RuntimeError("@@@ NEED TO IMPLEMENT!")

    def _r_catalog(self):
        raise RuntimeError("@@@ NEED TO IMPLEMENT!")



# debuggin
from lxml import etree

# 3rd-party modules
from lxml.builder import E

# module packages
from jnpr.junos.cfg import Resource
from jnpr.junos import jxml as JXML
from jnpr.junos.cfg.srx.shared_ab_addr import SharedAddrBookAddr
from jnpr.junos.cfg.srx.shared_ab_set import SharedAddrBookSet


class SharedAddrBook(Resource):

    """
    [edit security address-book <name>]

    Resource <name>
      The address book name, string

    Manages:
      addr - SharedAddrBookAddr resources
      set  - SharedAddrBookAddrSet resources
    """
    PROPERTIES = [
        'description',
        '$addrs',         # read-only addresss
        '$sets',          # read-only address-sets
        'zone_list'       # attached zone
    ]

    def __init__(self, junos, name=None, **kvargs):
        if name is None:
            # resource-manager
            Resource.__init__(self, junos, name, **kvargs)
            return

        self.addr = SharedAddrBookAddr(junos, parent=self)
        self.set = SharedAddrBookSet(junos, parent=self)
        self._manages = ['addr', 'set']
        Resource.__init__(self, junos, name, **kvargs)

    def _xml_at_top(self):
        return E.security(
            E('address-book', E.name(self._name))
        )

    # -----------------------------------------------------------------------
    # XML reading
    # -----------------------------------------------------------------------

    def _xml_hook_read_begin(self, xml):
        ab = xml.find('.//address-book')
        ab.append(E('description'))
        ab.append(E('address', JXML.NAMES_ONLY))
        ab.append(E('address-set', JXML.NAMES_ONLY))
        ab.append(E('attach'))
        return True

    def _xml_at_res(self, xml):
        return xml.find('.//address-book')

    def _xml_to_py(self, as_xml, to_py):
        Resource._r_has_xml_status(as_xml, to_py)
        Resource.copyifexists(as_xml, 'description', to_py)
        to_py['$addrs'] = [name.text for name in as_xml.xpath('address/name')]
        to_py['$sets'] = [
            name.text for name in as_xml.xpath('address-set/name')]

    # -----------------------------------------------------------------------
    # XML writing
    # -----------------------------------------------------------------------

    def _xml_change_zone_list(self, xml):
        x_attach = E('attach')
        self._xml_list_property_add_del_names(x_attach,
                                              prop_name='zone_list',
                                              element_name='zone')
        xml.append(x_attach)
        return True

    # -----------------------------------------------------------------------
    # Manager List, Catalog
    # -----------------------------------------------------------------------

    def _r_list(self):
        raise RuntimeError("Need to implement!")

    def _r_catalog(self):
        raise RuntimeError("Need to implement!")



from jnpr.junos.facts.chassis import chassis
from jnpr.junos.facts.routing_engines import routing_engines
from jnpr.junos.facts.personality import personality
from jnpr.junos.facts.swver import software_version
from jnpr.junos.facts.ifd_style import ifd_style
from jnpr.junos.facts.switch_style import switch_style
from jnpr.junos.facts.session import *
from jnpr.junos.facts.srx_cluster import srx_cluster

FACT_LIST = [
    chassis,
    routing_engines,
    personality,
    srx_cluster,
    software_version,
    ifd_style,
    switch_style,
    session
]

__all__ = ['FACT_LIST']



# 3rd-party modules
from lxml.builder import E

# module packages
from jnpr.junos.cfg import Resource
from jnpr.junos import jxml as JXML


class Application(Resource):

    """
    [edit applications application <name>]

    Resource name: str
      <name> is the application item name
    """

    PROPERTIES = [
        'description',
        'protocol',
        'dest_port',
        'timeout'
    ]

    # -----------------------------------------------------------------------
    # XML readers
    # -----------------------------------------------------------------------

    def _xml_at_top(self):
        return E.applications(E.application(E.name(self._name)))

    def _xml_at_res(self, xml):
        return xml.find('.//application')

    def _xml_to_py(self, has_xml, has_py):
        Resource._r_has_xml_status(has_xml, has_py)
        Resource.copyifexists(has_xml, 'description', has_py)
        has_py['protocol'] = has_xml.findtext('protocol')
        Resource.copyifexists(has_xml, 'destination-port', has_py, 'dest_port')
        Resource.copyifexists(has_xml, 'inactivity-timeout', has_py, 'timeout')
        if 'timeout' in has_py and has_py['timeout'] != 'never':
            has_py['timeout'] = int(has_py['timeout'])

    # -----------------------------------------------------------------------
    # XML property writers
    # -----------------------------------------------------------------------

    def _xml_change_protocol(self, xml):
        xml.append(E.protocol(self['protocol']))
        return True

    def _xml_change_dest_port(self, xml):
        """
        destination-port could be a single value or a range.
        handle the case where the value could be provided as either
        a single int value or a string range, e.g. "1-27"
        """
        value = self['dest_port']
        if isinstance(value, int):
            value = str(value)
        xml.append(E('destination-port', value))
        return True

    def _xml_change_timeout(self, xml):
        xml.append(E('inactivity-timeout', str(self['timeout'])))
        return True

    # -----------------------------------------------------------------------
    # Manager List, Catalog
    # -----------------------------------------------------------------------

    def _r_list(self):
        get = E.applications(E.application(JXML.NAMES_ONLY))
        got = self.R.get_config(get)
        self._rlist = [
            app.text for app in got.xpath('applications/application/name')]

    def _r_catalog(self):
        get = E.applications(E.application())
        got = self.R.get_config(get)
        for app in got.xpath('applications/application'):
            name = app.findtext('name')
            self._rcatalog[name] = {}
            self._xml_to_py(app, self._rcatalog[name])



from lxml import etree
from jnpr.junos import jxml


class RpcError(Exception):

    def __init__(self, cmd=None, rsp=None, errs=None):
        """
          :cmd: is the rpc command
          :rsp: is the rpc response (after <rpc-reply>)
          :errs: is a list of <rpc-error> elements
        """
        self.cmd = cmd
        self.rsp = rsp
        self.errs = errs

    def __repr__(self):
        """
          pprints the response XML attribute
        """
        if None != self.rsp:
            return etree.tostring(self.rsp, pretty_print=True)


class CommitError(RpcError):

    def __init__(self, cmd=None, rsp=None, errs=None):
        RpcError.__init__(self, cmd, rsp, errs)
        self.rpc_error = jxml.rpc_error(rsp)


class LockError(RpcError):

    def __init__(self, rsp):
        RpcError.__init__(self, rsp=rsp)
        self.rpc_error = jxml.rpc_error(rsp)


class UnlockError(RpcError):

    def __init__(self, rsp):
        RpcError.__init__(self, rsp=rsp)
        self.rpc_error = jxml.rpc_error(rsp)



# 3rd-party modules
from lxml.builder import E

# module packages
from jnpr.junos.cfg.srx.policy_rule import PolicyRule
from jnpr.junos.cfg import Resource
from jnpr.junos import jxml as JXML


class PolicyContext(Resource):

    """
    [edit security policy from-zone <from_zone> to-zone <to_zone>]

    Resource name: tuple(from_zone, to_zone)
      <from_zone>  is the name of the From zone
      <to_zone> is the name of the To zone

    Manages Resources:
      rule, PolicyRule
    """

    PROPERTIES = [
        '$rules',
        '$rules_count'
    ]

    def __init__(self, junos, name=None, **kvargs):
        if name is None:
            # resource-manager
            Resource.__init__(self, junos, name, **kvargs)
            return

        # specific instance will manage policy rules
        self.rule = PolicyRule(junos, M=self, parent=self)
        self._manages = ['rule']
        self._name_from_zone = name[0]
        self._name_to_zone = name[1]
        Resource.__init__(self, junos, name, **kvargs)

    def _xml_at_top(self):
        return E.security(E.policies(
            E.policy(
                E('from-zone-name', self._name_from_zone),
                E('to-zone-name', self._name_to_zone)
            )))

    # -------------------------------------------------------------------------
    # XML reading
    # -------------------------------------------------------------------------

    def _xml_config_read(self):
        """
          ~! OVERLOADS !~
        """
        xml = self._xml_at_top()
        xml.find('.//policy').append(E.policy(JXML.NAMES_ONLY))
        return self._junos.rpc.get_config(xml)

    def _xml_at_res(self, xml):
        return xml.find('.//policy')

    def _xml_to_py(self, as_xml, to_py):
        Resource._r_has_xml_status(as_xml, to_py)
        to_py['$rules'] = [
            policy.text for policy in as_xml.xpath('.//policy/name')]
        to_py['$rules_count'] = len(to_py['$rules'])

    # -----------------------------------------------------------------------
    # Resource List, Catalog
    # -- only executed by 'manager' resources
    # -----------------------------------------------------------------------

    def _r_list(self):
        """
          build the policy context resource list from the command:
          > show security policies zone-context
        """
        got = self._junos.rpc.get_firewall_policies(zone_context=True)

        for pc in got.xpath('//policy-zone-context/policy-zone-context-entry'):
            from_zone = pc.find('policy-zone-context-from-zone').text
            to_zone = pc.find('policy-zone-context-to-zone').text
            self._rlist.append((from_zone, to_zone))

    def _r_catalog(self):
        got = self._junos.rpc.get_firewall_policies(zone_context=True)

        for pc in got.xpath('//policy-zone-context/policy-zone-context-entry'):
            from_zone = pc.find('policy-zone-context-from-zone').text
            to_zone = pc.find('policy-zone-context-to-zone').text
            count = int(pc.find('policy-zone-context-policy-count').text)
            self._rcatalog[(from_zone, to_zone)] = {'$rules_count': count}



"""
facts['srx_cluster'] = True | False
  if personality.startswith('SRX')
  device responds to the "show" command to extract the data

Facts Required:
  personality
"""


def srx_cluster(junos, facts):
    if not facts['personality'].startswith('SRX'):
        return

    # we should check the 'cluster status' on redundancy group 0 to see who is
    # master.  we use a try/except block for cases when SRX is not clustered

    try:
        cluster_st = junos.rpc.get_chassis_cluster_status(redundancy_group="0")
        primary = cluster_st.xpath(
            './/redundancy-group-status[.="primary"]')[0]
        node = primary.xpath(
            'preceding-sibling::device-name[1]')[0].text.replace('node', 'RE')
        if node not in facts['master']:
            facts['master'].append(node)
        facts['srx_cluster'] = True
    except:
        facts['srx_cluster'] = False
        pass



# debuggin
from lxml import etree

# 3rd-party modules
from lxml.builder import E

# module packages
from jnpr.junos.cfg import Resource
from jnpr.junos import jxml as JXML
from jnpr.junos.cfg.srx.zone_ab import ZoneAddrBook
from jnpr.junos.cfg.srx.zone_ifs import ZoneInterface, HostInbSvcMixin


class Zone(HostInbSvcMixin, Resource):

    """
    [edit security zone security-zone <zone>]

    Resource Name: str
      The zone name

    Manages resources:
      ifs, ZoneInterface
      ab, ZoneAddressBook
    """

    PROPERTIES = [
        'description',
        'services',
        'protocols',
        '$ifs_list'
    ]

    def __init__(self, junos, name=None, **kvargs):
        if name is None:
            # resource-manager
            Resource.__init__(self, junos, name, **kvargs)
            return

        self.ifs = ZoneInterface(junos, parent=self)
        self.ab = ZoneAddrBook(junos, name, parent=self)
        self._manages = ['ifs', 'ab']
        Resource.__init__(self, junos, name, **kvargs)

    # -----------------------------------------------------------------------
    # XML reading
    # -----------------------------------------------------------------------

    def _xml_at_top(self):
        return E.security(E.zones(
            E('security-zone', E.name(self._name))
        ))

    def _xml_hook_read_begin(self, xml):
        e = xml.find('.//security-zone')
        e.append(E('host-inbound-traffic'))
        e.append(E('description'))
        e.append(E('interfaces'))
        return True

    def _xml_at_res(self, xml):
        return xml.find('.//security-zone')

    def _xml_to_py(self, as_xml, to_py):
        Resource._r_has_xml_status(as_xml, to_py)
        Resource.copyifexists(as_xml, 'description', to_py)
        e = as_xml.xpath('host-inbound-traffic/system-services/name')
        to_py['services'] = [n.text for n in e]
        e = as_xml.xpath('host-inbound-traffic/protocols/name')
        to_py['protocols'] = [n.text for n in e]
        to_py['$ifs_list'] = [
            name.text for name in as_xml.xpath('interfaces/name')]

    # -----------------------------------------------------------------------
    # XML writing
    # -----------------------------------------------------------------------

    # handled by mixin, yo!

    # -----------------------------------------------------------------------
    # Manager List, Catalog
    # -----------------------------------------------------------------------

    def _r_list(self):
        got = self.R.get_zones_information(terse=True)
        zones = got.findall('zones-security/zones-security-zonename')
        self._rlist = [zone.text for zone in zones]
        self._rlist.remove('junos-host')

    # using Resource._r_catalog()



# 3rd-party
from lxml.builder import E

# module packages
from jnpr.junos.cfg import Resource
from jnpr.junos import jxml as JXML


class ApplicationSet(Resource):

    """
    [edit applications application-set <name>]

    Resource name: str
      <name> is the application-set name
    """

    PROPERTIES = [
        'description',
        'app_list', 'app_list_adds', 'app_list_dels',
        'appset_list', 'appset_list_adds', 'appset_list_dels'
    ]

    def _xml_at_top(self):
        return E.applications(E('application-set', (E.name(self._name))))

    def _xml_at_res(self, xml):
        return xml.find('.//application-set')

    def _xml_to_py(self, has_xml, has_py):
        Resource._r_has_xml_status(has_xml, has_py)
        Resource.copyifexists(has_xml, 'description', has_py)

        # each of the <application> elements
        app_list = [this.findtext('name')
                    for this in has_xml.xpath('application')]
        set_list = [this.findtext('name')
                    for this in has_xml.xpath('application-set')]

        if len(app_list):
            has_py['app_list'] = app_list
        if len(set_list):
            has_py['appset_list'] = set_list

    # -----------------------------------------------------------------------
    # XML property writers
    # -----------------------------------------------------------------------

    # -------------------------------------------------------------------------
    # application list
    # -------------------------------------------------------------------------

    def _xml_change_app_list(self, xml):
        self._xml_list_property_add_del_names(xml,
                                              prop_name='app_list',
                                              element_name='application')
        return True

    def _xml_change_app_list_adds(self, xml):
        for this in self.should['app_list_adds']:
            xml.append(E.application(E.name(this)))
        return True

    def _xml_change_app_list_dels(self, xml):
        for this in self.should['app_list_dels']:
            xml.append(E.application(JXML.DEL, E.name(this)))
        return True

    # -------------------------------------------------------------------------
    # application-set list
    # -------------------------------------------------------------------------

    def _xml_change_appset_list(self, xml):
        if None == self.should.get('appset_list'):
            self['appset_list'] = []

        (adds, dels) = Resource.diff_list(
            self.has.get('appset_list', []), self.should['appset_list'])

        for this in adds:
            xml.append(E('application-set', E.name(this)))
        for this in dels:
            xml.append(E('application-set', JXML.DEL, E.name(this)))
        return True

    def _xml_change_appset_list_adds(self, xml):
        for this in self.should['appset_list_adds']:
            xml.append(E('application-set', E.name(this)))
        return True

    def _xml_change_appset_list_dels(self, xml):
        for this in self.should['appset_list_dels']:
            xml.append(E('application-set', JXML.DEL, E.name(this)))
        return True

    # -----------------------------------------------------------------------
    # Resource List, Catalog
    # -- only executed by 'manager' resources
    # -----------------------------------------------------------------------

    def _r_list(self):
        got = self.R.get_config(
            E.applications(E('application-set', JXML.NAMES_ONLY)))
        self._rlist = [this.text for this in got.xpath('.//name')]

    def _r_catalog(self):
        got = self.R.get_config(E.applications(E('application-set')))
        for this in got.xpath('.//application-set'):
            name = this.findtext('name')
            self._rcatalog[name] = {}
            self._xml_to_py(this, self._rcatalog[name])



from jnpr.junos.cfg.srx.app import Application
from jnpr.junos.cfg.srx.app_set import ApplicationSet
from jnpr.junos.cfg.srx.policy import PolicyContext
from jnpr.junos.cfg.srx.policy_rule import PolicyRule
from jnpr.junos.cfg.srx.zone import Zone
from jnpr.junos.cfg.srx.shared_ab import SharedAddrBook
from jnpr.junos.cfg.srx.zone_ab import ZoneAddrBook
from jnpr.junos.cfg.srx.ab_finder import AddrBookFinder



# 3rd-party packages
from lxml.builder import E

# local packages
from jnpr.junos.cfg import Resource
from jnpr.junos import jxml as JXML


class PolicyRule(Resource):

    """
    [edit security policy from-zone <from_zone> to-zone <to_zone> policy
        <policy_name>]

    Resource name: str
      <policy_name> is the name of the policy

    Managed by: PolicyContext
      <from_zone> and <to_zone> taken from parent resource
    """

    PROPERTIES = [
        'description',
        'match_srcs',
        'match_dsts',
        'match_apps',
        'action',
        'count',
        'log_init',
        'log_close'
    ]

    # -------------------------------------------------------------------------
    # XML reading
    # -------------------------------------------------------------------------

    def _xml_at_top(self):
        xml = self._parent._xml_at_top()
        xml.find('.//policy').append(E.policy(E.name(self._name)))
        return xml

    def _xml_at_res(self, xml):
        return xml.find('.//policy/policy')

    def _xml_to_py(self, as_xml, to_py):
        Resource._r_has_xml_status(as_xml, to_py)
        Resource.copyifexists(as_xml, 'description', to_py)

        x_match = as_xml.find('match')
        x_then = as_xml.find('then')

        # collect up the 'match' criteria

        to_py['match_srcs'] = [
            this.text for this in x_match.findall('source-address')]
        to_py['match_dsts'] = [
            this.text for this in x_match.findall('destination-address')]
        to_py['match_apps'] = [
            this.text for this in x_match.findall('application')]

        # collect up the 'then' criteria

        to_py['action'] = x_then.xpath('permit | reject | deny')[0].tag

        if x_then.find('count') is not None:
            to_py['count'] = True
        if x_then.find('log/session-init') is not None:
            to_py['log_init'] = True
        if x_then.find('log/session-close') is not None:
            to_py['log_close'] = True

    # -------------------------------------------------------------------------
    # XML writing
    # -------------------------------------------------------------------------

    def _xml_change_action(self, xml):
        xml.append(E.then(E(self.should['action'])))
        return True

    def _xml_change_count(self, xml):
        xml.append(E.then(
            Resource.xmltag_set_or_del('count', self.should['count'])
        ))
        return True

    def _xml_change_log_init(self, xml):
        xml.append(E.then(E.log(
            Resource.xmltag_set_or_del('session-init', self.should['log_init'])
        )))
        return True

    def _xml_change_log_close(self, xml):
        xml.append(E.then(E.log(
            Resource.xmltag_set_or_del(
                'session-close',
                self.should['log_close'])
        )))
        return True

    def _xml_change_match_srcs(self, xml):
        adds, dels = Resource.diff_list(
            self.has['match_srcs'], self.should['match_srcs'])

        if len(adds):
            x_match = E.match()
            xml.append(x_match)
            for this in adds:
                x_match.append(E('source-address', E.name(this)))

        if len(dels):
            x_match = E.match()
            xml.append(x_match)
            for this in dels:
                x_match.append(E('source-address', JXML.DEL, E.name(this)))

        return True

    def _xml_change_match_dsts(self, xml):
        adds, dels = Resource.diff_list(
            self.has['match_dsts'], self.should['match_dsts'])

        if len(adds):
            x_match = E.match()
            xml.append(x_match)
            for this in adds:
                x_match.append(E('destination-address', E.name(this)))

        if len(dels):
            x_match = E.match()
            xml.append(x_match)
            for this in dels:
                x_match.append(
                    E('destination-address', JXML.DEL, E.name(this)))

        return True

    def _xml_change_match_apps(self, xml):
        adds, dels = Resource.diff_list(
            self.has['match_apps'], self.should['match_apps'])

        if len(adds):
            x_match = E.match()
            xml.append(x_match)
            for this in adds:
                x_match.append(E('application', E.name(this)))

        if len(dels):
            x_match = E.match()
            xml.append(x_match)
            for this in dels:
                x_match.append(E('application', JXML.DEL, E.name(this)))

        return True

    # -----------------------------------------------------------------------
    # Resource List, Catalog
    # -- only executed by 'manager' resources
    # -----------------------------------------------------------------------

    def _r_list(self):
        got = self.P._xml_config_read()
        self._rlist = [
            this.text for this in got.xpath('.//policy/policy/name')]

    def _r_catalog(self):
        got = self.D.rpc.get_config(self.P._xml_at_top())
        policies = got.find('.//security/policies/policy')
        for pol in policies.findall('policy'):
            name = pol.find('name').text
            self._rcatalog[name] = {}
            self._xml_to_py(pol, self._rcatalog[name])



# 3rd-party modules
from lxml.builder import E

# module packages
from jnpr.junos.cfg import Resource
from jnpr.junos import jxml as JXML


class ZoneAddrBookAddr(Resource):

    """
    [edit security zone security-zone <zone> address-book address <name>]

    Resource name: str
      <name> is the name of the address item

    Managed by: ZoneAddrBook
      <zone> is the name of the security zone
    """

    PROPERTIES = [
        'description',
        'ip_prefix',
    ]

    def _xml_at_top(self):
        return E.security(E.zones(
            E('security-zone',
              E.name(self.P._name),
              E('address-book', E.address(self._name))
              )
        ))

    # -----------------------------------------------------------------------
    # XML reading
    # -----------------------------------------------------------------------

    def _xml_at_res(self, xml):
        return xml.find('.//address-book/address')

    def _xml_to_py(self, as_xml, to_py):
        Resource._r_has_xml_status(as_xml, to_py)
        Resource.copyifexists(as_xml, 'description', to_py)
        to_py['ip_prefix'] = as_xml.find('ip-prefix').text

    # -----------------------------------------------------------------------
    # XML writing
    # -----------------------------------------------------------------------

    def _xml_change_ip_prefix(self, xml):
        xml.append(E('ip-prefix', self.should['ip_prefix']))
        return True

    # -----------------------------------------------------------------------
    # Manager List, Catalog
    # -----------------------------------------------------------------------

    def _r_list(self):
        #  The parent keeps a property on this list, so just use it, yo!
        self._rlist = self.P['$addrs']

    def _r_catalog(self):
        get = E.security(E.zones(
            E('security-zone',
              E.name(self.P._name),
              E('address-book', E('address'))
              )
        ))
        got = self.D.rpc.get_config(get)
        for addr in got.xpath('.//address-book/address'):
            name = addr.find('name').text
            self._rcatalog[name] = {}
            self._xml_to_py(addr, self._rcatalog[name])



# debuggin
from lxml import etree

# 3rd-party modules
from lxml.builder import E

# module packages
from jnpr.junos.cfg import Resource
from jnpr.junos import jxml as JXML
from jnpr.junos.cfg.srx.zone_ab_addr import ZoneAddrBookAddr
from jnpr.junos.cfg.srx.zone_ab_set import ZoneAddrBookSet


class ZoneAddrBook(Resource):

    """
    [edit security zone security-zone <zone> address-book]

    Resource name: str
      The zone name

    Resource manages:
      addr, ZoneAddrBookAddr
      set, ZoneAddrBookSet
    """

    PROPERTIES = [
        '$addrs',         # read-only addresss
        '$sets'           # read-only address-sets
    ]

    def __init__(self, junos, name=None, **kvargs):
        if name is None:
            # resource-manager
            Resource.__init__(self, junos, name, **kvargs)
            return

        self.addr = ZoneAddrBookAddr(junos, parent=self)
        self.set = ZoneAddrBookSet(junos, parent=self)
        self._manages = ['addr', 'set']
        Resource.__init__(self, junos, name, **kvargs)

    def _xml_at_top(self):
        return E.security(E.zones(
            E('security-zone',
              E.name(self._name),
              E('address-book',
                E('address', JXML.NAMES_ONLY),
                E('address-set', JXML.NAMES_ONLY)
                )
              )
        ))

    # -----------------------------------------------------------------------
    # XML reading
    # -----------------------------------------------------------------------

    def _xml_at_res(self, xml):
        return xml.find('.//address-book')

    def _xml_to_py(self, as_xml, to_py):
        Resource._r_has_xml_status(as_xml, to_py)
        to_py['$addrs'] = [name.text for name in as_xml.xpath('address/name')]
        to_py['$sets'] = [
            name.text for name in as_xml.xpath('address-set/name')]

    # -----------------------------------------------------------------------
    # Manager List, Catalog
    # -----------------------------------------------------------------------

    def _r_list(self):
        # this list of zone addressbooks is really just the list of zones
        got = self.R.get_zones_information(terse=True)
        zones = got.findall('zones-security/zones-security-zonename')
        self._rlist = [zone.text for zone in zones]
        self._rlist.remove('junos-host')

    # using Resource._r_catalog()



# resources/srx/addrbook_finder.py
import netaddr

#from .zone import Zone
#from .addrbook import ZoneAddrBook


class AddrBookFinderResults(object):

    """
    Helper-class to hold the results of a :ZoneAddrFind.find(): invocation
    """

    def __init__(self, ab, find, results):
        self._ab = ab
        self._find = find
        self._results = results
        self.sets = []

    @property
    def lpm(self):
        """
        The longest-prefix-matching address is the last one in the results
        list. This fact is a result of the :ZoneAddrFinder.find(): sorted call
        """
        return self._results[-1][0]

    @property
    def items(self):
        """
        Return a list of the matching address items and sets
        """
        return self.addrs + self.sets

    @property
    def addrs(self):
        """
        Return a list of the matching address items
        """
        # return a list of names
        return [x[0] for x in self._results]

    @property
    def matching(self):
        """
        Returns the string value of the original querried address presented to
        the find() method
        """
        return self._find

    def __repr__(self):
        """
        Provides the matching value and the zone name associated with this
        results
        """
        return "%s(%s in %s)" % (
            self.__class__.__name__, self._find, self._ab.name)


class AddrBookFinder(object):

    # -------------------------------------------------------------------------
    # CONSTRUCTOR
    # -------------------------------------------------------------------------

    def __init__(self, addr_book):
        """
        addr_book
          Either a ZoneAddrBook or SharedAddrBook instance
        """
        self._ab = addr_book
        self._index = None

    def __repr__(self):
        return "AddrBookFinder(%s)" % self._ab.name

    def compile(self):
        """
        Compile a list of netaddr objects against the catalog of address items
        """
        # create a tuple of (addr-name, netaddr) for each of the items in the
        # address-book
        self._index = [(name, netaddr.IPNetwork(addr['ip_prefix']))
                       for name, addr in self._ab.addr.catalog.items()]

    def find(self, addr, sets=True):
        """
        Given an ip or ip_prefix locate the matching address book address
        and address-set items.
        """

        # if the caller hasn't explicity invoked :compile(): to create the
        # netaddr objects, then do that now.

        if self._index is None:
            self.compile()

        # convert the provided :addr: into a netaddr object and then
        # to a subnet match to find address entries.  the matching
        # values will be sorted with longest prefix matching to be
        # last in the list

        ip = netaddr.IPNetwork(addr).ip
        # is ip in the subnet?
        in_net = lambda i: ip & i[1].netmask == i[1].network
        # used to sort by prefix-length
        by_pflen = lambda a, b: cmp(a[1].prefixlen, b[1].prefixlen)
        r = sorted(
            filter(
                in_net,
                self._index),
            cmp=by_pflen)               # find/sort
        if r is None:
            return None

        # now that we have some matching entries, we should find which
        # address-set items uses the items

        results = AddrBookFinderResults(self._ab, addr, r)
        if sets is True:
            results.sets = self.find_sets(results)

        # return the results object
        return results

    def find_sets(self, r):
        """
        Given a :AddrBookFinderResults: object, which contains the list of
        matching address items, locate the list of address-set objects that
        use those items
        """
        catalog = self._ab.set.catalog
        in_addr = lambda i: i in v['addr_list']
        sets = [k for k, v in catalog.items() if filter(in_addr, r.addrs)]
        in_set = lambda i: i in v['set_list']
        subsets = [k for k, v in catalog.items() if filter(in_set, sets)]
        return sets + subsets



# debuggin
from lxml import etree

# 3rd-party modules
from lxml.builder import E

# module packages
from jnpr.junos.cfg import Resource
from jnpr.junos import jxml as JXML


class HostInbSvcMixin(object):

    def _xml_hook_build_change_begin(self, xml):
        """
        HostInbSvcMixin
        """
        def _default_empty(prop_name):
            if prop_name in self.should and self.should[prop_name] is None:
                self.should[prop_name] = []

        _default_empty('services')
        _default_empty('protocols')

    def _xml__change__hit(self, xml, prop, ele_name):
        (adds, dels) = Resource.diff_list(self.has[prop], self.should[prop])
        if not adds and not dels:
            return False

        hit = E('host-inbound-traffic')
        for this in adds:
            hit.append(E(ele_name, E.name(this)))
        for this in dels:
            hit.append(E(ele_name, JXML.DEL, E.name(this)))
        xml.append(hit)
        return True

    def _xml_change_services(self, xml):
        return self._xml__change__hit(xml, 'services', 'system-services')

    def _xml_change_protocols(self, xml):
        return self._xml__change__hit(xml, 'protocols', 'protocols')


class ZoneInterface(HostInbSvcMixin, Resource):

    """
    [edit security zone security-zone <zone> interfaces <name>]

    Resource name: str
      <name> is the interface name, e.g. 'ge-0/0/0.12'

    Managed by: Zone
      <zone> is the name of the security zone
    """

    PROPERTIES = [
        'services',
        'protocols',
    ]

    def _xml_at_top(self):
        return E.security(E.zones(
            E('security-zone',
              E.name(self.P._name),
              E.interfaces(E.name(self._name))
              )
        ))

    # -----------------------------------------------------------------------
    # XML reading
    # -----------------------------------------------------------------------

    def _xml_at_res(self, xml):
        return xml.find('.//interfaces')

    def _xml_to_py(self, as_xml, to_py):
        Resource._r_has_xml_status(as_xml, to_py)
        e = as_xml.xpath('host-inbound-traffic/system-services/name')
        to_py['services'] = [n.text for n in e]
        e = as_xml.xpath('host-inbound-traffic/protocols/name')
        to_py['protocols'] = [n.text for n in e]

    # -----------------------------------------------------------------------
    # XML writing
    # -----------------------------------------------------------------------

    # handled by the mixin

    # -----------------------------------------------------------------------
    # Manager List, Catalog
    # -----------------------------------------------------------------------

    def _r_list(self):
        self._rlist = self.P['$ifs_list']

    # using Resource._r_catalog()



# 3rd-party modules
from lxml.builder import E

# module packages
from jnpr.junos.cfg import Resource
from jnpr.junos import jxml as JXML


class SharedAddrBookSet(Resource):

    """
    [edit security address-book <ab_name> address-set <name>]

    ~! WARNING !~
    This resource is managed only as a child of the :ZoneAddrBook:
    resource.  Do not create a manager instance of this class directly
    """
    PROPERTIES = [
        'description',
        'addr_list',        # list of address items
        'set_list',         # sets can contain a list of sets
    ]

    def _xml_at_top(self):
        xml = self.P._xml_at_top()
        xml.find('.//address-book').append(
            E('address-set', E.name(self._name))
        )
        return xml

    # -----------------------------------------------------------------------
    # XML reading
    # -----------------------------------------------------------------------

    def _xml_at_res(self, xml):
        return xml.find('.//address-book/address-set')

    def _xml_to_py(self, as_xml, to_py):
        Resource._r_has_xml_status(as_xml, to_py)
        Resource.copyifexists(as_xml, 'description', to_py)
        to_py['addr_list'] = [
            name.text for name in as_xml.xpath('address/name')]
        to_py['set_list'] = [
            name.text for name in as_xml.xpath('address-set/name')]

    # -----------------------------------------------------------------------
    # XML writing
    # -----------------------------------------------------------------------

    def _xml_change_addr_list(self, xml):
        self._xml_list_property_add_del_names(xml,
                                              prop_name='addr_list',
                                              element_name='address')
        return True

    def _xml_change_set_list(self, xml):
        self._xml_list_property_add_del_names(xml,
                                              prop_name='set_list',
                                              element_name='address-set')
        return True

    # -----------------------------------------------------------------------
    # Manager List, Catalog
    # -----------------------------------------------------------------------

    def _r_list(self):
        # list of address-book address-sets.  this list is managed by the
        # parent object, so just use that, yo!
        self._rlist = self.P['$sets']

    def _r_catalog(self):
        get = self.P._xml_at_top()
        get.find('.//address-book').append(E('address-set'))
        got = self.D.rpc.get_config(get)
        for adrset in got.xpath('.//address-set'):
            name = adrset.find('name').text
            self._rcatalog[name] = {}
            self._xml_to_py(adrset, self._rcatalog[name])



# 3rd-party modules
from lxml.builder import E

# module packages
from jnpr.junos.cfg import Resource
from jnpr.junos import jxml as JXML
from jnpr.junos.cfg.srx.nat.nat_proxy_arp import NatProxyArp


class NatStaticRule(Resource):

    """
    [edit security nat static rule-set <ruleset_name> rule <rule_name>]

    Resource namevar:
      rule_name, string. The ruleset_name is obtained from the resource parent

    """

    PROPERTIES = [
        "description",
        "match_dst_addr",
        "match_dst_port",
        "nat_addr",
        "nat_port",
        "proxy_interface"
    ]

    # -----------------------------------------------------------------------
    # XML read
    # -----------------------------------------------------------------------

    def _xml_at_top(self):
        return E.security(E.nat(E.static(
            E('rule-set',
              E.name(self.P._name),
              E.rule(E.name(self._name))
              )
        )))

    def _xml_at_res(self, xml):
        return xml.find('.//rule')

    def _xml_to_py(self, as_xml, to_py):
        """
        converts Junos XML to native Python
        """
        Resource._r_has_xml_status(as_xml, to_py)
        Resource.copyifexists(as_xml, 'description', to_py)
        e = as_xml.find('static-nat-rule-match')
        to_py['match_dst_addr'] = e.find('destination-address').text

    # -----------------------------------------------------------------------
    # XML write
    # -----------------------------------------------------------------------

    def _xml_hook_build_change_begin(self, xml):
        if 'nat_port' not in self.should:
            # if 'nat_port' is not provided, then default to the
            # 'match_dst_port' value
            self.should['nat_port'] = self['match_dst_port']

        if 'match_dst_addr' in self.should and 'proxy_interface' in self.has:
            # if we are changing the 'match_dst_addr' and we also have a proxy
            # interface, then we need to update the proxy_interface value to
            # the new match_dst_addr value. start by deleting the existing one:
            namevar = (self['proxy_interface'], self.has['match_dst_addr'])
            NatProxyArp(self._junos, namevar).delete()

            if 'proxy_interface' not in self.should:
                # if the 'proxy_interface' value was not actually changed, then
                # simply copy the current one into :should:  this will trigger
                # the flush/create in the property-writer below
                self.should['proxy_interface'] = self.has['proxy_interface']

        # build up some XML that will be used by the property-writers
        match = E('static-nat-rule-match')
        xml.append(match)
        then = E.then(E('static-nat', E('prefix')))
        xml.append(then)
        self._rxml_match = match
        self._rxml_then = then.find('static-nat/prefix')

    def _xml_change_match_dst_addr(self, xml):
        self._rxml_match.append(
            E('destination-address',
              JXML.REPLACE,
              self.should['match_dst_addr'])
        )
        return True

    def _xml_change_match_dst_port(self, xml):
        self._rxml_match.append(
            E('destination-port', E.low(self.should['match_dst_port']))
        )
        return True

    def _xml_change_nat_addr(self, xml):
        self._rxml_then.append(E('addr-prefix', self.should['nat_addr']))
        return True

    def _xml_change_nat_port(self, xml):
        self._rxml_then.append(
            E('mapped-port', E('low', self.should['nat_port'])))
        return True

    def _xml_change_proxy_interface(self, xml):
        # this is really always going to be a 'create a new resource'. If the
        # caller is changing the 'match_dst_addr' value, then the existing
        # entry will be removed by the "hook" function.
        namevar = (self.should['proxy_interface'], self['match_dst_addr'])
        parp = NatProxyArp(self._junos, namevar)
        parp.write(touch=True)
        return True

    # -----------------------------------------------------------------------
    # Manager List, Catalog
    # -----------------------------------------------------------------------

    def _r_list(self):
        self._rlist = self.P['rules']

    def _r_catalog(self):
        get = E.security(E.nat(E.static(
            E('rule-set',
              E.name(self.P._name),
              )
        )))
        got = self.D.rpc.get_config(get)
        for rule in got.xpath('.//rule'):
            name = rule.find('name').text
            self._rcatalog[name] = {}
            self._xml_to_py(rule, self._rcatalog[name])



# 3rd-party modules
from lxml.builder import E

# module packages
from jnpr.junos.cfg import Resource
from jnpr.junos import jxml as JXML
from jnpr.junos.cfg.srx.nat.nat_st_rule import NatStaticRule


class NatStaticRuleSet(Resource):

    """
    [edit security nat static rule-set <name>]
    """

    PROPERTIES = [
        "description",
        "zone_from",
        "$rules",
        "$rules_count"
    ]

    def __init__(self, junos, name=None, **kvargs):
        if name is None:
            # resource-manager
            Resource.__init__(self, junos, name, **kvargs)
            return

        self.rule = NatStaticRule(junos, M=self, parent=self)
        self._manages = ['rule']
        Resource.__init__(self, junos, name, **kvargs)

    # -----------------------------------------------------------------------
    # XML read
    # -----------------------------------------------------------------------

    def _xml_at_top(self):
        return E.security(E.nat(E.static(
            E('rule-set', E.name(self._name))
        )))

    def _xml_hook_read_begin(self, xml):
        """
        need to add the <from>,<to> elements to pick up the zone context
        need to add the rules, names-only
        """
        rs = xml.find('.//rule-set')
        rs.append(E('from'))
        rs.append(E.rule(JXML.NAMES_ONLY))
        return True

    def _xml_at_res(self, xml):
        return xml.find('.//rule-set')

    def _xml_to_py(self, as_xml, to_py):
        Resource._r_has_xml_status(as_xml, to_py)
        Resource.copyifexists(as_xml, 'description', to_py)
        to_py['zone_from'] = as_xml.find('from/zone').text
        to_py['$rules'] = [rule.text for rule in as_xml.xpath('.//rule/name')]
        to_py['$rules_count'] = len(to_py['$rules'])

    # -----------------------------------------------------------------------
    # XML write
    # -----------------------------------------------------------------------

    def _xml_change_zone_from(self, xml):
        xml.append(E('from', JXML.REPLACE, E.zone(self.should['zone_from'])))
        return True

    # -----------------------------------------------------------------------
    # Resource List, Catalog
    # -- only executed by 'manager' resources
    # -----------------------------------------------------------------------

    def _r_list(self):
        get = E.security(E.nat(E.static(
            E('rule-set', JXML.NAMES_ONLY)
        )))
        got = self.D.rpc.get_config(get)
        self._rlist = [name.text for name in got.xpath('.//name')]

    def _r_catalog(self):
        get = E.security(E.nat(E.static(
            E('rule-set')
        )))
        got = self.D.rpc.get_config(get)
        for ruleset in got.xpath('.//rule-set'):
            name = ruleset.find("name").text
            self._rcatalog[name] = {}
            self._xml_to_py(ruleset, self._rcatalog[name])



# 3rd-party modules
from lxml.builder import E

# module packages
from jnpr.junos.cfg import Resource
from jnpr.junos import jxml as JXML


class NatSrcRule(Resource):

    """
    [edit security nat source rule-set <ruleset-name> rule <rule-name>]
    """

    PROPERTIES = [
        "match_src_addr",
        "match_dst_addr",
        "pool"
    ]

    # -----------------------------------------------------------------------
    # XML read
    # -----------------------------------------------------------------------

    def _xml_at_top(self):
        return E.security(E.nat(E.source(
            E('rule-set',
              E.name(self.P._name),
              E.rule(E.name(self._name))
              )
        )))

    def _xml_at_res(self, xml):
        return xml.find('.//rule')

    def _xml_to_py(self, as_xml, to_py):
        """
        converts Junos XML to native Python
        """
        Resource._r_has_xml_status(as_xml, to_py)
        e = as_xml.find('src-nat-rule-match')
        to_py['match_src_addr'] = e.find('source-address').text
        to_py['match_dst_addr'] = e.find('destination-address').text
        to_py['pool'] = as_xml.find('.//pool-name').text

    # -----------------------------------------------------------------------
    # XML write
    # -----------------------------------------------------------------------

    def _xml_hook_build_change_begin(self, xml):
        """
        when doing a write, assign default values if they are not present
        """
        def _default_to(prop, value):
            if prop not in self.should:
                self.should[prop] = value

        if self.is_new:
            _default_to('match_dst_addr', '0.0.0.0/0')
            _default_to('match_src_addr', '0.0.0.0/0')

    def _xml_change_match_src_addr(self, xml):
        xml.append(E('src-nat-rule-match',
                     E('source-address',
                       JXML.REPLACE,
                       self.should['match_src_addr'])
                     ))
        return True

    def _xml_change_match_dst_addr(self, xml):
        xml.append(E('src-nat-rule-match',
                     E('destination-address',
                       JXML.REPLACE,
                       self.should['match_dst_addr'])
                     ))
        return True

    def _xml_change_pool(self, xml):
        xml.append(E.then(
            E('source-nat', E.pool(E('pool-name', self.should['pool'])))
        ))
        return True

    # -----------------------------------------------------------------------
    # Manager List, Catalog
    # -----------------------------------------------------------------------

    def _r_list(self):
        self._rlist = self.P['$rules']

    def _r_catalog(self):
        get = E.security(E.nat(E.source(
            E('rule-set',
              E.name(self.P._name),
              )
        )))
        got = self.D.rpc.get_config(get)
        for rule in got.xpath('.//rule'):
            name = rule.find('name').text
            self._rcatalog[name] = {}
            self._xml_to_py(rule, self._rcatalog[name])



# 3rd-party modules
from lxml.builder import E

# module packages
from jnpr.junos.cfg import Resource
from jnpr.junos import jxml as JXML


class NatSrcPool(Resource):

    """
    [edit security nat source pool <name>]
    """

    PROPERTIES = [
        'addr_from',
        'addr_to'
    ]

    def _xml_at_top(self):
        """
        configuration to retrieve resource
        """
        return E.security(E.nat(E.source(E.pool(E.name(self.name)))))

    # -----------------------------------------------------------------------
    # XML read
    # -----------------------------------------------------------------------

    def _xml_at_res(self, xml):
        """
        return Element at resource
        """
        return xml.find('.//pool')

    def _xml_to_py(self, as_xml, to_py):
        """
        converts Junos XML to native Python
        """
        Resource._r_has_xml_status(as_xml, to_py)
        to_py['addr_from'] = as_xml.find('address/name').text
        to_py['addr_to'] = as_xml.find('address/to/ipaddr').text

    # -----------------------------------------------------------------------
    # XML property writers
    # -----------------------------------------------------------------------

    def _xml_change_addr_from(self, xml):
        # we need to always set the address/name given the structure of the
        # Junos configuration XML, derp.

        addr_from = self.should.get('addr_from') or self.has.get('addr_from')
        xml.append(E.address(JXML.REPLACE, E.name(addr_from)))
        return True

    def _xml_change_addr_to(self, xml):
        # we must always include the addr_from, so if we didn't expliclity
        # change it, we must do it now.

        if 'addr_from' not in self.should:
            self._xml_change_addr_from(xml)

        x_addr = xml.find('address')
        x_addr.append(E.to(E.ipaddr(self.should['addr_to'])))
        return True

    # -----------------------------------------------------------------------
    # Resource List, Catalog
    # -- only executed by 'manager' resources
    # -----------------------------------------------------------------------

    def _r_list(self):
        """
          build the policy context resource list from the command~
          > show security policies zone-context
        """
        get = E.security(E.nat(E.source(
            E.pool(JXML.NAMES_ONLY))))

        got = self.D.rpc.get_config(get)
        self._rlist = [name.text for name in got.xpath('.//pool/name')]

    def _r_catalog(self):
        get = E.security(E.nat(E.source(E.pool)))
        got = self.D.rpc.get_config(get)
        for pool in got.xpath('.//pool'):
            name = pool.find('name').text
            self._rcatalog[name] = {}
            self._xml_to_py(pool, self._rcatalog[name])



from jnpr.junos.cfg.srx.nat.nat_src_pool import NatSrcPool
from jnpr.junos.cfg.srx.nat.nat_src_ruleset import NatSrcRuleSet
from jnpr.junos.cfg.srx.nat.nat_src_rule import NatSrcRule

from jnpr.junos.cfg.srx.nat.nat_st_ruleset import NatStaticRuleSet
from jnpr.junos.cfg.srx.nat.nat_st_rule import NatStaticRule
from jnpr.junos.cfg.srx.nat.nat_proxy_arp import NatProxyArp



from lxml.builder import E
from jnpr.junos import jxml as JXML


def chassis(junos, facts):
    """
    Obtain basic chassis facts:
      model : product model
      serialnumber : serial number
      hostname : host
      fqdn : fully-qualified domain-name
      domain : domain-name

    NOTES:
    (1) if in a 2RE system, this routine will only load the information
        from the first chassis item.
    (2) hostname, domain, and fqdn are retrieved from configuration data;
        inherited configs are checked.
    """
    rsp = junos.rpc.get_chassis_inventory()

    if rsp.tag == 'multi-routing-engine-results':
        facts['2RE'] = True
        x_ch = rsp.xpath('.//chassis-inventory')[0].find('chassis')
    else:
        facts['2RE'] = False
        x_ch = rsp.find('chassis')

    facts['model'] = x_ch.find('description').text
    try:
        facts['serialnumber'] = x_ch.find('serial-number').text
    except:
        # if the toplevel chassis does not have a serial-number, then
        # check the Backplane chassis-module
        facts['serialnumber'] = x_ch.xpath(
            'chassis-module[name="Backplane"]/serial-number')[0].text

    got = junos.rpc.get_config(
        E.system(
            E('host-name'),
            E('domain-name')
        ),
        JXML.INHERIT
    )

    hostname = got.find('.//host-name')
    facts['hostname'] = hostname.text if hostname is not None else 'Amnesiac'
    facts['fqdn'] = facts['hostname']

    domain = got.find('.//domain-name')
    if domain is not None:
        facts['domain'] = domain.text
        facts['fqdn'] += '.%s' % facts['domain']
    else:
        facts['domain'] = None
